package org.processmining.plugins.dcnet;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.geom.AffineTransform;
import java.awt.geom.Arc2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Line2D.Double;
import java.awt.geom.Path2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.RectangularShape;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;

import javax.swing.JPopupMenu;

import org.processmining.models.cnet.CNetBinding;
import org.processmining.models.cnet.CNetNode;
import org.processmining.plugins.dcnet.NodeDecorator.BorderStyle;
import org.processmining.plugins.graphviz.dot.DotNode;

import com.kitfox.svg.SVGDiagram;

final class BindingNode {

	static final double DOUBLE_BORDER_FACTOR = 1.4;

	public interface SelectionListener {
		void deselected(BindingNode node, SVGDiagram image);

		void selected(BindingNode node, SVGDiagram image);
	}

	public static final class DefaultBindingNodeSelectionListener implements SelectionListener {

		public void selected(BindingNode node, SVGDiagram image) {
			node.setSelected(true);
		}

		public void deselected(BindingNode node, SVGDiagram image) {
			node.setSelected(false);
		}
	}

	private final CNetNode task;
	private final CNetBinding binding;
	private final BindingType type;

	private final NodeDecorator decorator;

	/**
	 * all in original SVG coordinates
	 */
	private final Collection<Connector> connectors;

	/**
	 * in original SVG coordinates
	 */
	private final RectangularShape boundingShape;

	private Shape connectorShape;

	private PopupDrawer popupDrawer = new PopupDrawer();

	private boolean isSelectable = true;
	private boolean isSelected = false;
	private double scaleFactor = 1.2;

	private List<MouseListener> mouseListener = new CopyOnWriteArrayList<>();
	private List<SelectionListener> selectionListener = new CopyOnWriteArrayList<>();
	private String label;

	public BindingNode(CNetNode task, final CNetBinding binding, BindingType type, DotNode dotNode,
			RectangularShape boundingShape, List<Connector> connectors, final NodeDecorator decorator) {
		this.task = task;
		this.binding = binding;
		this.decorator = decorator;
		this.label = decorator.getBindingLabel(binding);
		this.type = type;
		this.boundingShape = boundingShape;
		this.connectors = connectors;
		double connectorSize = decorator.getBindingSize(binding);
		this.connectorShape = new Ellipse2D.Double(-(connectorSize / 2.0), -(connectorSize / 2.0), connectorSize,
				connectorSize);
		mouseListener.add(new MouseAdapter() {

			@Override
			public void mousePressed(MouseEvent e) {
				if (e.isPopupTrigger()) {
					showMenu(e);
				}
			}

			@Override
			public void mouseReleased(MouseEvent e) {
				if (e.isPopupTrigger()) {
					showMenu(e);
				}
			}

			private void showMenu(MouseEvent e) {
				JPopupMenu menu = decorator.getBindingMenu(binding);
				if (menu != null && menu.getComponentCount() > 0) {
					menu.show(e.getComponent(), e.getX(), e.getY());
				}
			}

		});
	}

	public void decorateBinding(SVGDiagram svg, Graphics2D g) {
		Color oldColor = g.getColor();
		Stroke oldStroke = g.getStroke();

		Iterator<Connector> iterator = getConnectors().iterator();

		if (connectors.size() == 1) {
			paintShape(g, iterator.next());
		} else {
			Shape shape = getBoundingShape();
			if (shape instanceof Rectangle2D) {
				// Draw connectors
				Connector connA = iterator.next();
				paintShape(g, connA);
				// Draw lines between all points
				g.setStroke(getLineStroke());
				Path2D.Double path = new Path2D.Double();
				Point2D posA = connA.getPosition();
				path.moveTo(posA.getX(), posA.getY());
				while (iterator.hasNext()) {
					Connector connB = iterator.next();
					paintShape(g, connB);
					g.setStroke(getLineStroke());
					Point2D posB = connB.getPosition();
					path.lineTo(posB.getX(), posB.getY());
					connA = connB;
				}
				g.draw(path);
			} else if (shape instanceof Ellipse2D) {
				// Draw connectors
				Connector lastPoint = iterator.next();
				paintShape(g, lastPoint);
				while (iterator.hasNext()) {
					Connector nextPoint = iterator.next();
					paintShape(g, nextPoint);
					g.setStroke(getLineStroke());
					if (nextPoint.isCorrectPosition() && lastPoint.isCorrectPosition()) {
						// Draw arc 
						g.draw(createArc(lastPoint, nextPoint, (Ellipse2D) shape));
					} else {
						g.draw(createLine(lastPoint, nextPoint));
					}
				}
			}
		}

		g.setColor(oldColor);
		g.setStroke(oldStroke);
	}

	private void paintShape(Graphics2D g, Connector connector) {
		Shape shape = createPort(connector);
		if (decorator.getBindingBorderStyle(binding) != BorderStyle.NONE) {
			drawBorder(g, createPortBorder(connector));
		}
		fillShape(g, shape);
		drawShape(g, shape);
		g.setColor(decorator.getBindingEdgeColor(binding));
	}

	private void fillShape(Graphics2D g, Shape shape) {
		g.setColor(decorator.getBindingFillColor(binding));
		g.fill(shape);
	}

	private void drawShape(Graphics2D g, Shape shape) {
		Stroke stroke = decorator.getBindingStroke(binding);
		if (stroke != null) {
			g.setStroke(stroke);
			g.setColor(decorator.getBindingStrokeColor(binding));
			g.draw(shape);
		}
	}

	private void drawBorder(Graphics2D g, Shape borderShape) {
		Stroke stroke = decorator.getBindingBorderStroke(binding);
		if (stroke != null) {
			g.setStroke(stroke);
			g.setColor(decorator.getBindingBorderColor(binding));
			g.draw(borderShape);
			g.setColor(Color.WHITE);
			g.fill(borderShape);			
		}
	}

	public void decoratedSelectedBinding(SVGDiagram svg, Graphics2D g) {

		if (isSelected()) {

			Color oldColor = g.getColor();
			Stroke oldStroke = g.getStroke();
			Font oldFont = g.getFont();

			double x = Integer.MIN_VALUE;
			double y = Integer.MIN_VALUE;

			for (Connector connector : connectors) {
				x = Math.max(connector.getPosition().getX(), x);
				y = Math.max(connector.getPosition().getY(), y);
			}

			if (label != null && !label.isEmpty()) {
				popupDrawer.drawLabelPopup(label, g, (float) x, (float) y);
			}

			// DEBUG code
			//g.setColor(Color.RED);
			//g.draw(getBoundingShape());
			//g.setColor(Color.BLACK);

			g.setColor(oldColor);
			g.setStroke(oldStroke);
			g.setFont(oldFont);
		}

	}

	private Double createLine(Connector lastPoint, Connector nextPoint) {
		Point2D posA = lastPoint.getPosition();
		Point2D posB = nextPoint.getPosition();
		return new Line2D.Double(posA.getX(), posA.getY(), posB.getX(), posB.getY());
	}

	private Arc2D.Double createArc(Connector connA, Connector connB, Ellipse2D ellipse) {
		Point2D pointA = connA.getPosition();
		Point2D pointB = connB.getPosition();
		double sx = pointA.getX() - ellipse.getCenterX();
		double sy = pointA.getY() - ellipse.getCenterY();
		double rx = ellipse.getWidth() / 2.0d;
		double ry = ellipse.getHeight() / 2.0d;
		double start = EllipseUtils.degreeOnEllipse(sx, sy, rx, ry);
		double ex = pointB.getX() - ellipse.getCenterX();
		double ey = pointB.getY() - ellipse.getCenterY();
		double end = EllipseUtils.degreeOnEllipse(ex, ey, rx, ry);

		//TODO simplify if possible
		double extent = (end - start);
		if (extent > 180) {
			extent = -(360 - extent);
		} else if (extent < -180) {
			extent = (extent + 360);
		}
		Arc2D.Double arc = new Arc2D.Double(ellipse.getBounds2D(), start, extent, Arc2D.OPEN);
		return arc;
	}

	private Shape createPort(Connector connector) {
		return AffineTransform.getTranslateInstance(connector.getPosition().getX(), connector.getPosition().getY())
				.createTransformedShape(getConnectorShape());
	}

	private Shape createPortBorder(Connector connector) {
		AffineTransform borderTransform = new AffineTransform();
		if (decorator.getBindingBorderStyle(binding) == BorderStyle.DOUBLE) {
			borderTransform = AffineTransform.getScaleInstance(DOUBLE_BORDER_FACTOR, 1.4);
		}
		Shape scaledShape = borderTransform.createTransformedShape(getConnectorShape());
		return AffineTransform.getTranslateInstance(connector.getPosition().getX(), connector.getPosition().getY())
				.createTransformedShape(scaledShape);
	}

	public Collection<Connector> getConnectors() {
		return connectors;
	}

	public BindingType getType() {
		return type;
	}

	public CNetNode getTask() {
		return task;
	}

	public CNetBinding getBinding() {
		return binding;
	}

	public boolean isSelectable() {
		return isSelectable;
	}

	public List<MouseListener> getMouseListeners() {
		return mouseListener;
	}

	public List<SelectionListener> getSelectionListeners() {
		return selectionListener;
	}

	public boolean isSelected() {
		return isSelected;
	}

	public void setSelected(boolean isSelected) {
		this.isSelected = isSelected;
	}

	public Stroke getLineStroke() {
		if (isSelected) {
			return new BasicStroke((float) (decorator.getBindingEdgeWidth(binding) * scaleFactor), BasicStroke.CAP_BUTT,
					BasicStroke.JOIN_ROUND, 10.0f, new float[] { 2.0f }, 0.0f);
		} else {
			return new BasicStroke(decorator.getBindingEdgeWidth(binding), BasicStroke.CAP_BUTT,
					BasicStroke.JOIN_ROUND);
		}
	}

	public Shape getConnectorShape() {
		if (isSelected) {
			return AffineTransform.getScaleInstance(scaleFactor, scaleFactor).createTransformedShape(connectorShape);
		} else {
			return connectorShape;
		}
	}

	public RectangularShape getBoundingShape() {
		return boundingShape;
	}

}