package org.processmining.plugins.dcnet;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Deque;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import org.processmining.models.cnet.CNetNode;
import org.processmining.models.cnet.CausalNet;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.Ordering;

import edu.uci.ics.jung.graph.DirectedGraph;
import edu.uci.ics.jung.graph.DirectedSparseGraph;

public final class FeedbackArcUtil {

	private FeedbackArcUtil() {
	}

	public static Set<CNetRelation> approximateFeedbackArcSet(CausalNet cnet) {

		DirectedGraph<CNetNode, CNetRelation> graph = new DirectedSparseGraph<>();
		for (CNetNode node : cnet.getNodes()) {
			graph.addVertex(node);
			for (CNetNode successor : cnet.getSuccessors(node)) {
				CNetRelation relation = CNetRelation.Factory.create(node, successor);
				if (!graph.containsEdge(relation)) {
					graph.addEdge(relation, node, successor);
				}
			}
			for (CNetNode predecessor : cnet.getPredecessors(node)) {
				CNetRelation relation = CNetRelation.Factory.create(predecessor, node);
				if (!graph.containsEdge(relation)) {
					graph.addEdge(relation, predecessor, node);
				}
			}
		}

		Deque<CNetNode> s1 = new ArrayDeque<>();
		Deque<CNetNode> s2 = new ArrayDeque<>();
		Set<CNetNode> sinks = new LinkedHashSet<>();
		Set<CNetNode> sources = new LinkedHashSet<>();
		while (graph.getVertexCount() > 0) {
			for (CNetNode node : sortedVertices(graph)) {
				if (graph.outDegree(node) == 0) {
					sinks.add(node);
				}
			}
			while (!sinks.isEmpty()) {
				Iterator<CNetNode> iterator = sinks.iterator();
				CNetNode sink = iterator.next();
				iterator.remove();
				s2.push(sink);
				graph.removeVertex(sink);
				if (sinks.isEmpty()) {
					for (CNetNode node : sortedVertices(graph)) {
						if (graph.outDegree(node) == 0) {
							sinks.add(node);
						}
					}
				}
			}
			for (CNetNode node : sortedVertices(graph)) {
				if (graph.inDegree(node) == 0) {
					sources.add(node);
				}
			}
			while (!sources.isEmpty()) {
				Iterator<CNetNode> iterator = sources.iterator();
				CNetNode source = iterator.next();
				iterator.remove();
				s1.add(source);
				graph.removeVertex(source);
				if (sources.isEmpty()) {
					for (CNetNode node : graph.getVertices()) {
						if (graph.inDegree(node) == 0) {
							sources.add(node);
						}
					}
				}
			}
			if (graph.getVertexCount() > 0) {
				CNetNode maxNode = null;
				int maxDelta = Integer.MIN_VALUE;
				for (CNetNode node : sortedVertices(graph)) {
					int inDegree = graph.inDegree(node);
					int outDegree = graph.outDegree(node);
					int delta = outDegree - inDegree;
					if (delta > maxDelta) {
						maxDelta = delta;
						maxNode = node;
					}
				}
				s1.add(maxNode);
				graph.removeVertex(maxNode);
			}
		}

		List<CNetNode> vertexSeq = new ArrayList<>(s1);
		vertexSeq.addAll(s2);
		Set<CNetNode> visited = new HashSet<>();
		Set<CNetRelation> feedbackArcs = new HashSet<>();

		for (CNetNode node : vertexSeq) {
			visited.add(node);
			for (CNetNode suc : cnet.getSuccessors(node)) {
				if (visited.contains(suc)) {
					feedbackArcs.add(CNetRelation.Factory.create(node, suc));
				}
			}
		}

		return feedbackArcs;
	}

	private static ImmutableList<CNetNode> sortedVertices(DirectedGraph<CNetNode, CNetRelation> graph) {
		return Ordering.natural().immutableSortedCopy(graph.getVertices());
	}

}
